import os

import paramiko
import yaml

from si_tests import settings
from si_tests.deployments.utils import (kubectl_utils, commons,
                                        file_utils, wait_utils)
from si_tests.deployments.utils.namespace import NAMESPACE
from si_tests.managers.openstack_client_manager import OpenStackClientManager


def deploy_refapp(openstack_client_manager: OpenStackClientManager, kubeconfig: str = None):
    """
    Deploy RefApp heat stack to OpenStack Cluster
    Args:
        openstack_client_manager: OpenStack manager
        kubeconfig: cluster KUBECONFIG or os.environ.get("KUBECONFIG") if None

    Returns: None

    """

    stack_name = "refapp-test"
    source = file_utils.join(commons.REFAPP_DIR, "heat-templates")
    if not os.path.exists(source):
        commons.LOG.warning("repapp source does not exist %s", source)

    client_name = openstack_client_manager.client.name
    kubectl = kubectl_utils.Kubectl(kubeconfig=kubeconfig)
    destination = "{}/{}:{}".format(
        NAMESPACE.openstack, client_name, "/tmp/heat-templates")

    commons.LOG.info("Copy heat templates")
    kubectl.cp(source, destination)

    commons.LOG.info("Create heat stack")
    openstack_client_manager.stack.create(
        [stack_name, "-t", "/tmp/heat-templates/top.yaml"])

    commons.LOG.info("Wait for heat stack to create")
    wait = wait_utils.Waiter(openstack_client_manager.os_manager, 3600)

    def f():
        status = openstack_client_manager.stack.show(
            [stack_name])["stack_status"]
        if status == "CREATE_FAILED":
            resource_list = openstack_client_manager.stack.resource_list(
                ['-n', '10', stack_name])
            event_list = openstack_client_manager.stack.event_list(
                ['--nested-depth', '10', stack_name])
            commons.LOG.error("Resource info: %s\nHeat stack event list: %s",
                              resource_list, event_list)
            raise Exception("Failed to create stack")
        return status == "CREATE_COMPLETE"

    wait.wait(f)

    commons.LOG.info("Refapp successfully deployed")


def collect_logs_refapp(openstack_client_manager, proxy_ip=None, dst_dir=settings.ARTIFACTS_DIR):
    """KUBECONFIG is required"""

    if (proxy_ip):
        proxy_connection = commons.get_ssh_client(proxy_ip)
    elif (not settings.get_var_as_bool('RUN_ON_REMOTE', False)):
        proxy_ip = settings.SEED_STANDALONE_EXTERNAL_IP
        if (proxy_ip):
            proxy_connection = commons.get_ssh_client(ip=proxy_ip, auth=commons.get_ssh_auth(
                username=settings.SEED_SSH_LOGIN,
                key_filename=settings.SEED_SSH_PRIV_KEY_FILE))
        else:
            proxy_ip = settings.PHYS_SEED_FIXED_IP
            if (proxy_ip):
                proxy_connection = commons.get_ssh_client(proxy_ip)
            else:
                proxy_ip = openstack_client_manager.os_manager.api._K8sCluster__api_client.configuration.host
                if (proxy_ip):
                    proxy_ip = proxy_ip.split(':')[1][2:]
                    proxy_connection = commons.get_ssh_client(proxy_ip)
                else:
                    commons.LOG.error("No proxy IP for access to reffapp stack")
                    return

    stack_name = "refapp-test"
    stack_info = openstack_client_manager.stack.show([stack_name])
    if (len(stack_info) > 1):
        vm_list = openstack_client_manager.stack.resource_list(
            ['-n', '10', stack_name, '-c', 'physical_resource_id', '--filter',
             'type=OS::Nova::Server', '-f', 'value']).split()
        if vm_list:
            commons.LOG.info("Collect logs from RefApp stack")
            work_dir = os.path.join(dst_dir, stack_name + ".log")
            os.mkdir(work_dir)
            instances_dir = os.path.join(work_dir, "instances")
            os.mkdir(instances_dir)
            with open(os.path.join(work_dir, "stack_info.yaml"), "w") as f:
                yaml.safe_dump(stack_info, f)
            lb_list = openstack_client_manager.stack.resource_list(
                ['-n', '10', stack_name, '-c', 'physical_resource_id', '--filter',
                 'type=OS::Octavia::LoadBalancer', '-f', 'value']).split()
            if lb_list:
                current_dir = os.path.join(work_dir, "loadbalancer")
                os.mkdir(current_dir)
                for lb_item in lb_list:
                    lb_info = openstack_client_manager.loadbalancer.show([lb_item])
                    with open(os.path.join(current_dir, lb_info["name"] + ".yaml"), "w") as f:
                        yaml.safe_dump(lb_info, f)
                    amphora_list = openstack_client_manager.loadbalancer.amphora_list(
                        ['--loadbalancer', lb_item, '-c', 'id', '-f', 'value']).split()
                    if amphora_list:
                        for amphora_item in amphora_list:
                            amphora_info = openstack_client_manager.loadbalancer.amphora_show([amphora_item])
                            with open(os.path.join(current_dir, "amphora-" + amphora_item + ".yaml"), "w") as f:
                                yaml.safe_dump(amphora_info, f)
                            if amphora_info["compute_id"] is not None:
                                vm_list.append(amphora_info["compute_id"])
            for instance_id in vm_list:
                instance_info = openstack_client_manager.server.show([instance_id])
                if instance_info and isinstance(instance_info, dict):
                    current_dir = os.path.join(instances_dir, instance_info["name"])
                    os.mkdir(current_dir)
                    with open(os.path.join(current_dir, "instance_info.yaml"), "w") as f:
                        yaml.safe_dump(instance_info, f)
                    try:
                        instance_ip_list = openstack_client_manager.server_get_floating_ips(instance_id)
                        if instance_ip_list:
                            if (proxy_connection):
                                ssh_connect = proxy_connection.proxy_to(host=instance_ip_list[0],
                                                                        auth=commons.get_ssh_auth())
                            else:
                                ssh_connect = commons.get_ssh_client(host=instance_ip_list[0])
                            openstack_client_manager.server_collect_logs(ssh_connect=ssh_connect,
                                                                         arch_path=current_dir,
                                                                         docker_logs=True)
                        else:
                            openstack_client_manager.server_collect_console(instance_id,
                                                                            arch_path=current_dir)
                    except (paramiko.ssh_exception.SSHException):
                        openstack_client_manager.server_collect_console(instance_id,
                                                                        arch_path=current_dir)
    else:
        commons.LOG.info("Stack %s does not exist. Nothing collect.", stack_name)


def delete_refapp(openstack_client_manager):
    """
    Delete RefApp heat stack from the OpenStack Cluster
    Args:
        openstack_client_manager: OpenStack manager

    Returns: None

    """
    """KUBECONFIG is required"""

    stack_name = "refapp-test"

    commons.LOG.info("Delete RefApp stack")
    openstack_client_manager.stack.delete([stack_name])

    wait = wait_utils.Waiter(openstack_client_manager.os_manager, 3600)

    def wait_refapp_stack_delete():
        response = openstack_client_manager.stack.show(
            [stack_name])
        return 'Stack not found' in response

    commons.LOG.info("Wait for RefApp heat stack to delete")
    wait.wait(wait_refapp_stack_delete)

    commons.LOG.info("Refapp heat stack successfully deleted")
