from si_tests import logger
from si_tests import settings
from si_tests.managers.kaas_manager import Manager, Machine
from si_tests.utils.utils import generate_list_pods
from si_tests.utils.utils import NodeLabel
import pytest
import os
import logging
from tabulate import tabulate
from dataclasses import dataclass
LOG = logger.logger
init_parent_handler = LOG.parent.handlers[0]

LOG.info("=========GENERATING POD LIST=========")
generated_pod_list = generate_list_pods(
    kaas_manager=Manager(kubeconfig=settings.KUBECONFIG_PATH),
    target_nss=settings.HA_TEST_NAMESPACE_LIST
)


@pytest.fixture()
def cluster_condition_check(target_cluster):
    LOG.info("Check that all expected pods and "
             "their replicas are presented before HA test")
    try:
        target_cluster.check.check_machines_status()
        target_cluster.check.check_cluster_nodes()
        target_cluster.check.check_k8s_nodes()
        target_cluster.check.check_k8s_pods()
        target_cluster.check.check_actual_expected_pods()
    except Exception as e:
        LOG.error(f"Cluster {target_cluster.name} is not ready for HA test")
        LOG.debug(e)
        raise

    yield
    LOG.info("Check that all expected pods and "
             "their replicas are presented after HA test")
    try:
        target_cluster.check.check_machines_status()
        target_cluster.check.check_cluster_nodes()
        target_cluster.check.check_k8s_nodes()
        target_cluster.check.check_k8s_pods()
        target_cluster.check.check_actual_expected_pods()
    except Exception as e:
        LOG.error(f"Cluster {target_cluster.name} is not ready after HA test")
        LOG.debug(e)
        raise


@pytest.fixture()
def helmbundles_check(target_cluster):
    LOG.info("Check that helmbundles ready before HA test")
    try:
        target_cluster.check.check_helmbundles()
        target_cluster.check.check_cluster_readiness()
    except Exception as e:
        LOG.error(f"Cluster {target_cluster.name} is not ready for HA test")
        LOG.debug(e)
        raise

    yield
    LOG.info("Check that helmbundles ready after HA test")
    try:
        target_cluster.check.check_helmbundles()
        target_cluster.check.check_cluster_readiness()
    except Exception as e:
        LOG.error(f"Cluster {target_cluster.name} is not ready after HA test")
        LOG.debug(e)
        raise


@pytest.fixture(scope='function', params=generated_pod_list.values(),
                ids=generated_pod_list.keys())
def pod_group(request, target_cluster):
    """This fixture parametrize tests by group of pods
    that will be used in each test scenario.
    After scenario is finished, fixture will check phases
    for all pods in cluster (should be running)
    """
    # fetching variables
    failed_num = request.session.testsfailed
    pod_name, pod_num, pods = request.param

    try:
        # replace LOG handler for the test
        LOG.parent.handlers.clear()
        log_name = request.node.name.split('[')[0] + '-' + pod_name
        log_file = os.path.join(settings.LOGS_DIR,
                                log_name + ".log")
        fh = logging.FileHandler(log_file)
        fh.setFormatter(LOG.handlers[0].formatter)
        LOG.parent.addHandler(fh)

        # Compare actual and expected pods
        assert len(pods) != 0, "Cannot find {} running pods".format(pod_name)
        if not pod_num:
            # new pods are discovered that not specified in expected pod list
            # we do not skip them for now
            LOG.error("Pod ({0}) is not defined in "
                      "expected list".format(pod_name))
        if len(pods) != int(pod_num):
            # we can fail test here, but for now it is error in logs
            LOG.error("Mismatch between expected ({0}) and "
                      "actual number ({1}) for {2} "
                      "running pods".format(pod_num, len(pods), pod_name))

        nss = target_cluster.expected_pods.keys()
        # Check cluster health
        target_cluster.check.check_k8s_nodes(timeout=20, interval=2)
        target_cluster.check.check_k8s_pods(timeout=170, interval=10,
                                            target_namespaces=nss)
    except Exception as e:
        LOG.parent.handlers.clear()
        LOG.parent.addHandler(init_parent_handler)
        raise e
    yield request.param
    # replace LOG handler back
    LOG.parent.handlers.clear()
    LOG.parent.addHandler(init_parent_handler)
    if failed_num == request.session.testsfailed:
        # remove log file if test passed
        os.remove(log_file)


commands = {'kill_all': 'pkill -u $(whoami)',
            'kill_all_except_init': 'kill -9 -1'}


@pytest.fixture(scope='module', params=commands.values(),
                ids=list(commands.keys()))
def command(request):
    yield request.param


@dataclass
class MachineLabelData:
    machine: Machine
    node_name: str
    machine_name: str
    all_labels: NodeLabel or None
    all_labels_raw: list


@dataclass
class TargetMachineData:
    labels: list
    machine: MachineLabelData


def collect_machines_labels():
    """
    Collect all labels from machines and k8s nodes

    Returns: NodeData objects list
    """
    result = []
    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    managed_ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    cluster = managed_ns.get_cluster(settings.TARGET_CLUSTER)

    # find ceph_osd machines
    ceph_cluster = cluster.get_kaascephcluster()
    ceph_cluster_status = ceph_cluster.data.get('status') or {}
    osd_mapping = ceph_cluster_status.get('fullClusterInfo', {}).get('cephDetails', {}).get(
        'deviceMapping', {}).keys()
    ceph_osd_machines = [*osd_mapping]
    # find keepalive machine
    keepalive_machine = cluster.get_keepalive_master_machine()

    for machine in cluster.get_machines():
        machine_labels = machine.metadata['labels']
        k8s_node = machine.get_k8s_node().read()
        k8s_node_name = machine.get_k8s_node_name()
        k8s_node_labels = k8s_node.metadata.labels
        all_labels_raw = [f"{k}={v}" for k, v in {**machine_labels, **k8s_node_labels}.items()]
        all_labels = set()
        for item in NodeLabel:
            if item.value in ";".join(all_labels_raw):
                all_labels.add(item)
        if NodeLabel.kaas_master not in all_labels:
            # Workaround for missing labels for node role on worker nodes
            all_labels.add(NodeLabel.kaas_worker)
        if machine.name in ceph_osd_machines:
            # add ceph_osd label from kaascephcluster data
            all_labels.add(NodeLabel.ceph_osd)
        if machine.name == keepalive_machine.name:
            all_labels.add(NodeLabel.kaas_keepalive)
        result.append(MachineLabelData(machine=machine,
                                       machine_name=machine.name,
                                       node_name=k8s_node_name,
                                       all_labels=all_labels or None,
                                       all_labels_raw=all_labels_raw))
    return result


def collect_machine_sequence():
    """
    Generate machine sequences by expected HA_NODE_REBOOT_LABELS

    Returns: TargetMachine objects list

    """
    if not settings.HA_NODE_REBOOT_LABELS:
        raise Exception("Please specify expected node labels in HA_NODE_REBOOT_LABELS env variable")
    result = []
    actual_labels = collect_machines_labels()
    expected_labels = [NodeLabel.get_by_name(x) for x in settings.HA_NODE_REBOOT_LABELS]
    # check if expected labels exist in target machines
    for item in expected_labels:
        if not [x for x in actual_labels if item in x.all_labels]:
            # raise exception if no machine found with expected label
            raise Exception(f"Cannot find any node with label: '{item.name}'")
    for item in actual_labels:
        labels = [i for i in expected_labels if i in item.all_labels]
        if labels:
            result.append(TargetMachineData(labels=labels, machine=item))

    res_table = {item.machine_name: [label.name for label in item.all_labels] for item in actual_labels}
    LOG.info("Expected labels: {}".format(", ".join([label.name for label in expected_labels])))
    LOG.info('Actual node labels:\n' + tabulate(res_table.items(), headers=["Node", "Labels"], tablefmt="presto"))
    return result
