#    Copyright 2022 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
import json
from datetime import datetime, timedelta

import pytest
from tabulate import tabulate

from si_tests import logger, settings
from si_tests.deployments import deploy_refapp
from si_tests.managers.bootstrap_manager import BootstrapManager
from si_tests.managers.load_tools_manager import LoadToolsManager
from si_tests.managers.openstack_client_manager import OpenStackClientManager

LOG = logger.logger


def _calculate_downtime(data: dict) -> list:
    """
    Calculate potential downtime period

    Args:
        data: LoadToolsManager.get_custom_statistics object

    Returns: result list of downtimes

    """
    result = []
    start = {}
    stop = {}
    for item in data:
        if item['fail_per_sec']:
            if not start:
                start = item
            else:
                stop = item
        else:
            if start and stop:
                time_format = '%H:%M:%S'
                start_data = datetime.strptime(start['time_data'], time_format)
                stop_data = datetime.strptime(stop['time_data'], time_format)
                if stop_data < start_data:
                    # workaround to handle next day because only time is stored
                    stop_data += timedelta(days=1)
                duration = stop_data - start_data

                result.append({
                    "start": start_data.strftime(time_format),
                    "stop": stop_data.strftime(time_format),
                    "duration": duration
                })
            start = {}
            stop = {}
    return result


def _check_downtime(statistics, expected_downtime: int = settings.REFAPP_EXPECTED_DOWNTIME):
    """
    Check is existing downtime in list more than expected
    Args:
        expected_downtime: time in seconds
        statistics: LoadToolsManager.get_custom_statistics object

    Returns: None

    """
    downtime = _calculate_downtime(statistics)
    LOG.error("The following downtime ranges(s) found:\n" + tabulate(downtime, tablefmt="presto", headers="keys"))
    not_expected_downtime = list(filter(lambda x: x['duration'] > timedelta(seconds=expected_downtime), downtime))
    if not_expected_downtime:
        LOG.error(f"The following downtime ranges(s) more that expected ({expected_downtime} s.)\n"
                  + tabulate(not_expected_downtime, tablefmt="presto", headers="keys"))
        raise Exception("RefApp downtime during testing is longer than expected")
    else:
        LOG.info("ReffApp downtime within expected limits")


@pytest.fixture(scope='function')
def loadtest_os_refapp(kaas_manager):
    """
    Deploy RefApp to OpenStack Cluster (if not exist) as a workload

    """
    if settings.ALLOW_WORKLOAD:
        LOG.info("Start RefApp load")
        managed_ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
        cluster = managed_ns.get_cluster(settings.TARGET_CLUSTER)

        child_kubeconfig_name, child_kubeconfig = \
            cluster.get_kubeconfig_from_secret()
        with open(child_kubeconfig_name, 'w') as f:
            f.write(child_kubeconfig)
        cluster_os_manager = OpenStackClientManager(kubeconfig=child_kubeconfig_name)

        if cluster_os_manager.stack.check(["refapp-test"]).startswith("Stack not found"):
            LOG.info("Deploy RefApp")
            deploy_refapp.deploy_refapp(cluster_os_manager, kubeconfig=child_kubeconfig_name)
        else:
            LOG.info("RefApp already installed")

        refapp_url = cluster_os_manager.refapp_url
        LOG.info(f"RefApp URL = {refapp_url}")

        if settings.SEED_STANDALONE_EXTERNAL_IP:
            seed_ip = settings.SEED_STANDALONE_EXTERNAL_IP
        else:
            seed_ip = kaas_manager.si_config.data.get('run_on_remote', {}).get('SEED_STANDALONE_EXTERNAL_IP')
        assert seed_ip, 'Seed node IP address not found'
        LOG.info(f"Seed node IP address {seed_ip}")
        bootstrap_manager = BootstrapManager(seed_ip=seed_ip)
        load_manager = LoadToolsManager(bootstrap_manager=bootstrap_manager, refapp_address=refapp_url)
        load_manager.run_loadtools_on_seed_node()
        load_manager.start_swarm()

        yield

        try:
            datetime_prefix = datetime.now().strftime("%Y%m%d%H%M%S")
            load_manager.stop_swarm()
            statistics = load_manager.get_custom_statistics()

            html_report_filename = f'report-{cluster.name}-{datetime_prefix}.html'
            json_statistic_filename = f"statistics-{cluster.name}-{datetime_prefix}.json"

            load_manager.receive_report(file_name=html_report_filename)

            with open(f'{settings.ARTIFACTS_DIR}{json_statistic_filename}', 'w') as fp:
                json.dump(statistics, fp)

            LOG.info(f"Save HTML report to artifacts/{html_report_filename}")
            LOG.info(f"Save JSON statistic to artifacts/{json_statistic_filename}")

            LOG.info("Check ReffApp downtime")
            _check_downtime(statistics=statistics)
        except Exception as e:
            raise e
        finally:
            load_manager.exit()
    else:
        LOG.info("Skip RefApp load")

        yield
