#    Copyright 2022 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import pytest

from si_tests import logger
from si_tests import settings
from si_tests.managers import bootstrap_manager
from si_tests.managers.kaas_manager import Manager
from si_tests.clients.openstack import OpenstackClient

LOG = logger.logger


class TestProvisionRegionalClusters():

    @pytest.mark.usefixtures("introspect_openstack_seed_regional_deploy_objects")
    def test_regional_bootstrapv2(self, kaas_manager: Manager,
                                  openstack_client: OpenstackClient, show_step):
        """Deploy Bootstrap V2 Regional cluster on KaaS mgmt

        Scenario:
            1. Check if a provider is eligible to start
            2. Prepare templates to apply on bootstrap cluster
            2. Start bootstrap V2 regional cluster deployment
            3. Check result of deployment
        """

        # step 001: Check if a provider is eligible to start
        show_step(1)
        if not (settings.KAAS_AWS_ENABLED or
                settings.KAAS_AZURE_ENABLED or
                settings.KAAS_VSPHERE_ENABLED or
                settings.KAAS_OPENSTACK_ENABLED):
            pytest.fail("Unknown provider for bootstrapv2")

        # step 002: Prepare templates to apply on bootstrap cluster
        show_step(2)
        bootstrap = bootstrap_manager.BootstrapManager(seed_ip=openstack_client.get_seed_ip())
        bootstrap.step_003_prepare_seed_node_templates()
        bootstrap.step_003_patch_bootstrapv2_release_refs()

        # TODO: (morgoev) check if kind cluster has bootstrapregions
        # if so drop and recreate the cluster

        # step 003: Start bootstrap V2 regional cluster deployment
        show_step(3)
        endpoints = None
        if settings.KAAS_OPENSTACK_ENABLED:
            endpoints = openstack_client.get_os_cloud_endpoints()
        # NOTE: region isn't required in bv2 scenario
        bootstrap.step_004_deploy_kaas_regional(region="", endpoints=endpoints)

        # step 004: Check result of deployment
        show_step(4)
        bootstrap.step_006_wait_for_pods()
        bootstrap.step_006_postdeployment_checks()

        ns = kaas_manager.get_namespace(settings.REGION_NAMESPACE)
        regional_cluster = ns.get_cluster(settings.REGIONAL_CLUSTER_NAME)
        regional_cluster.check.check_machines_status()
        regional_cluster.check.check_cluster_nodes(timeout=900)
        regional_cluster.check.check_helmbundles()
        regional_cluster.check.check_persistent_volumes_mounts()

        regional_cluster.store_k8s_artifacts()

        if settings.OPENSTACK_BOOT_FROM_VOLUME:
            bootstrap.check_boot_from_volume(openstack_client,
                                             boot_volume_size=settings.OPENSTACK_BOOT_VOLUME_SIZE)

        # Check/wait for correct docker service replicas in cluster
        regional_cluster.check.check_actual_expected_docker_services()
        regional_cluster.check.check_k8s_pods()
        check_all_namespaces = False
        # NOTE: for some reason there is no check for cluster as well as
        # pods across all namespaces on azure atm in v1 scenario
        if not settings.KAAS_AZURE_ENABLED:
            regional_cluster.check.check_cluster_readiness()
            check_all_namespaces = True
        regional_cluster.check.check_actual_expected_pods(timeout=1200, check_all_nss=check_all_namespaces)
        regional_cluster.check.check_deploy_stage_success()
        regional_cluster.provider_resources.save_artifact()
