#    Copyright 2022 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pytest

from si_tests import logger
from si_tests import settings
from si_tests.fixtures.ha import collect_machine_sequence, TargetMachineData
from si_tests.managers.kaas_manager import Machine, Cluster, Namespace
from si_tests.managers.openstack_manager import OpenStackManager
from si_tests.utils import waiters

LOG = logger.logger


def id_func(param):
    return 'LABEL={};NODE={}'.format(";".join(map(lambda x: x.name, param.labels)), param.machine.machine_name)


@pytest.mark.usefixtures('loadtest_os_refapp')
@pytest.mark.parametrize("target_machine_data", collect_machine_sequence(), ids=id_func)
def test_ha_abnormal_reboot_mosk(request, kaas_manager, show_step, target_machine_data: TargetMachineData):
    """Abnormal shutdown machine in MOSK cluster

    Scenario:
            1. Abnormal reboot the node
            2. Wait for k8s to reboot
            3. Check cluster readiness
            4. Check Ceph cluster
            5. Check OpenStack resources

    Precondition:
            - Launch RefApp
            - Launch loadtest
    Postcondition:
            - Check workload downtime

    """
    if request.session.testsfailed:
        pytest.skip("Skip because only one node can be damage")

    managed_ns: Namespace = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    cluster: Cluster = managed_ns.get_cluster(settings.TARGET_CLUSTER)
    machine: Machine = target_machine_data.machine.machine
    k8s_node_name = machine.get_k8s_node_name()
    machine_ip = machine.public_ip

    show_step(1)
    machine.run_cmd("sudo reboot", check_exit_code=False)

    show_step(2)
    LOG.info("Wait until machine is no longer available via ICMP")
    waiters.wait(lambda: not waiters.icmp_ping(machine_ip), interval=5, timeout=600)

    LOG.info("Wait until k8s node become NotReady")
    cluster.check.wait_k8s_node_status(k8s_node_name, expected_status='NotReady', timeout=900)

    LOG.info("Wait until machine is available via ICMP")
    waiters.wait(lambda: waiters.icmp_ping(machine_ip), interval=5, timeout=1800)

    LOG.info("Wait until k8s node become Ready")
    cluster.check.wait_k8s_node_status(k8s_node_name, expected_status='Ready', timeout=1800)

    show_step(3)
    cluster.check.check_k8s_nodes(timeout=1800)
    # Check/wait for correct docker service replicas in cluster
    ucp_worker_agent_name = cluster.check.get_ucp_worker_agent_name()
    cluster.check.check_actual_expected_docker_services(
        changed_after_upd={'ucp-worker-agent-x': ucp_worker_agent_name})
    cluster.check.check_actual_expected_pods(timeout=3200)
    cluster.check.check_k8s_pods()
    cluster.check.check_cluster_readiness()

    show_step(4)
    cluster.check.wait_ceph_health_status(timeout=1800, interval=30)

    show_step(5)
    child_kubeconfig_name, child_kubeconfig = cluster.get_kubeconfig_from_secret()
    with open(child_kubeconfig_name, 'w') as f:
        f.write(child_kubeconfig)
    os_manager = OpenStackManager(kubeconfig=child_kubeconfig_name)
    LOG.info("Wait osdpl health status=Ready")
    os_manager.wait_openstackdeployment_health_status(timeout=1800)
    LOG.info("Wait os jobs to success and pods to become Ready")
    os_manager.wait_os_resources(timeout=1800)
