import pytest
from packaging import version
from kubernetes.client.rest import ApiException

from si_tests import logger
from si_tests.utils import waiters

LOG = logger.logger


class TestTFAdmissionController(object):
    """Test TFOperator admission controller"""
    helmbundle_name = "tungstenfabric-operator"

    @pytest.fixture(scope='class', autouse=True)
    def restore_helmbundle(self, os_manager):
        LOG.info("Backup helmbundle specs")
        spec_backup = os_manager.get_tf_helmbundle(
            self.helmbundle_name).data['spec']
        yield
        LOG.info("Restore helmbundle specs")
        os_manager.get_tf_helmbundle(self.helmbundle_name).patch(
            {"spec": spec_backup})

    @pytest.fixture(autouse=True)
    def setup_cls(self, setup, enable_admission_ctl):
        pass

    @pytest.fixture()
    def setup(self, os_manager):
        self.validating_webhook = f"{self.helmbundle_name}-validating-" \
                                  f"webhook-configuration"
        self.mutating_webhook = f"{self.helmbundle_name}-mutating-webhook-" \
                                f"configuration"
        self.tfoperator_version = os_manager.tfoperator_version()
        self.tf_helmbundle = os_manager.get_tf_helmbundle(self.helmbundle_name)
        self.tf_deployment = os_manager.get_tfoperator_deployment()
        self.tfoperator_cr = os_manager.get_tfoperator_cr()
        self.api_admission = os_manager.api.api_admission

    @pytest.fixture()
    def enable_admission_ctl(self):
        LOG.info("Enable admission controller")
        _, changed = self.set_admission_ctl(enabled=True)
        if changed:
            gen = self.tf_deployment.read().status.observed_generation
            waiters.wait(self.check_deployment, timeout=180,
                         predicate_args=[gen])

    @pytest.fixture()
    def restore_cr(self, os_manager):
        LOG.info("Backup tfoperator CR specs")
        tfoperator_cr = os_manager.get_tfoperator_cr()
        spec_backup = tfoperator_cr.data['spec']
        yield
        LOG.info("Restore tfoperator CR specs")
        tfoperator_cr.patch({"spec": spec_backup})

    def set_admission_ctl(self, enabled=True):
        spec = self.tf_helmbundle.data['spec']
        changed = False
        chart_name = 'tungstenfabric-operator'
        for chart in spec['releases']:
            if chart['name'] == chart_name:
                values = chart['values']
                if values.get('admission', {}).get('enabled', True) != enabled:
                    changed = True
                values['admission'] = {'enabled': enabled}
                return self.tf_helmbundle.patch({"spec": spec}), changed
        raise Exception("Chart '{}' not found.".format(chart_name))

    def check_deployment(self, gen):
        return self.tf_deployment.read().status.observed_generation > gen

    @staticmethod
    def check_webhook(func, *args, expected=True):
        try:
            func(*args)
            LOG.info("Webhook is present")
            return True and expected
        except ApiException as e:
            if e.status == 404:
                LOG.info("Webhook not found")
                return not expected

    def check_validating_webhook(self, expected=True):
        LOG.info(f'Check Validating Webhook {self.validating_webhook}')
        return self.check_webhook(
            self.api_admission.read_validating_webhook_configuration,
            self.validating_webhook, expected=expected,
        )

    def check_mutating_webhook(self, expected=True):
        LOG.info(f'Check Mutating Webhook {self.mutating_webhook}')
        return self.check_webhook(
            self.api_admission.read_mutating_webhook_configuration,
            self.mutating_webhook, expected=expected,
        )

    @pytest.mark.usefixtures("restore_cr")
    def test_invalidation_request(self, show_step):
        """Check valid and non-valid requests.

            Scenario:
                1. Check validation webhook
                2. Send and check invalid request
        """

        show_step(1)
        assert self.check_validating_webhook(expected=True), \
            "Validating Admission Webhook isn't created"

        show_step(2)
        invalid_settings = {
            "settings": {
                "tfImage": {
                    "tfImageTag": "5.1.20220101011200"
                },
                "tfVersion": "21.4",
            }
        }
        with pytest.raises(ApiException) as exc_info:
            self.tfoperator_cr.patch({"spec": invalid_settings})
        assert exc_info.value.status == 403
        assert exc_info.value.reason == 'Forbidden'

    @pytest.mark.usefixtures("restore_cr")
    def test_mutation_webhook(self, show_step):
        """Check mutating webhook.

            Scenario:
                1. Check mutating webhook
                2. Set previous TF version
                3. Check TF version was changed to default
        """
        if version.parse(self.tfoperator_version) <= version.parse("0.11"):
            pytest.skip("TFOperator version doesn't supported")
        default_version = "21.4"
        show_step(1)
        assert self.check_mutating_webhook(expected=True), \
            "Mutating Admission Webhook isn't created"

        show_step(2)
        settings = {
            "settings": {
                "tfVersion": "2011",
            }
        }
        self.tfoperator_cr.patch({"spec": settings})

        show_step(3)
        self.tfoperator_cr.read()
        assert self.tfoperator_cr.data['spec']['settings']['tfVersion'] == \
               default_version

    def test_manage_admission_controller(self, show_step):
        """Enable/disable tf-operator admission controller via helmbundle.

            Scenario:
                1. Disable admission controller
                2. Check Validating/Mutating Webhooks Configuration were
                removed
                3. Enable admission controller
                4. Check Validating/Mutating Webhooks Configuration were created
        """
        show_step(1)
        gen = self.tf_deployment.read().status.observed_generation
        self.set_admission_ctl(enabled=False)
        # Check deployment has been updated
        waiters.wait(self.check_deployment, timeout=180,
                     predicate_args=[gen])

        show_step(2)
        waiters.wait(self.check_validating_webhook, timeout=180,
                     predicate_kwargs={'expected': False})
        waiters.wait(self.check_mutating_webhook, timeout=180,
                     predicate_kwargs={'expected': False})

        show_step(3)
        gen = self.tf_deployment.read().status.observed_generation
        self.set_admission_ctl(enabled=True)
        # Check deployment has been updated
        waiters.wait(self.check_deployment, predicate_args=[gen])

        show_step(4)
        waiters.wait(self.check_validating_webhook, timeout=180,
                     predicate_kwargs={'expected': True})
        waiters.wait(self.check_mutating_webhook, timeout=180,
                     predicate_kwargs={'expected': True})
