#    Copyright 2022 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from cryptography import x509
from si_tests import logger
from si_tests import settings
from si_tests.managers import bootstrap_manager
from si_tests.utils import utils
from si_tests.utils import certs
from si_tests.utils import waiters

LOG = logger.logger


def test_set_child_custom_cert(kaas_manager, show_step):
    """Add custom cert to the child cluster.

    Scenario:
        1. Collecting env data
        2. Get seed ip and ssh-Key
        3. Get old cert from lb and kubeconfig data
        4. Generate and set new cert
        5. Wait for new cert on lb host
        6. Wait for update data in kubeconfig - WR for PRODX-27128
        7. Check and validate cert

    """

    def get_cert(h):
        LOG.info("Get new cert from host")
        return certs.get_certificate_from_host(hostname=h)

    def get_cert_data_from_kubeconfig():
        LOG.info("Get kubeconfig from secret")
        kc = kaas_manager.get_secret_data(namespace=namespace_name,
                                          secret_name=secret_name,
                                          data_key="admin.conf")
        kubeconfig = yaml.safe_load(kc)

        user_cert_data = kubeconfig['users'][0]['user']['client-certificate-data']
        cert_auth_data = kubeconfig['clusters'][0]['cluster']['certificate-authority-data']
        client_cert_data = kubeconfig['users'][0]['user']['client-key-data']
        return user_cert_data, cert_auth_data, client_cert_data

    def is_kubeconfig_changed(user_cert_data, cert_auth_data, client_cert_data):
        new_user_cert_data, new_cert_auth_data, new_client_cert_data = get_cert_data_from_kubeconfig()
        if any([new_user_cert_data == user_cert_data,
                new_cert_auth_data == cert_auth_data,
                new_client_cert_data == client_cert_data]):
            LOG.info("Kubeconfig data not changed yet")
            return False
        else:
            LOG.info("Kubeconfig data changed")
            return True

    show_step(1)
    cluster_name = settings.TARGET_CLUSTER
    namespace_name = settings.TARGET_NAMESPACE
    ns = kaas_manager.get_namespace(namespace_name)
    cluster = ns.get_cluster(cluster_name)
    show_step(2)
    seed_ip = kaas_manager.si_config.data.get('run_on_remote', {}).get('SEED_STANDALONE_EXTERNAL_IP')
    assert seed_ip, 'Seed node IP address not found'
    LOG.info(f"Seed node IP address {seed_ip}")
    key = kaas_manager.si_config.data['run_on_remote']['SEED_SSH_PRIV_KEY']
    with open(
            'key-file',
            mode='w') as f:
        f.write(key)
    bootstrap = bootstrap_manager.BootstrapManager(seed_ip=seed_ip, priv_key_file='key-file')
    host = cluster.load_balancer_ip
    ip = None
    dns_name = None
    if utils.is_valid_ip(host):
        ip = host
    else:
        dns_name = host
    show_step(3)
    cluster_data = cluster.get_lcm_cluster(name=cluster_name, namespace=namespace_name).read()
    control_nodes = cluster_data.spec['machineTypes']['control']
    deploy_params = [x for x in control_nodes if x['name'] == 'deploy'][0]['params']

    # Note
    # ucp_kube_apiserver_port used on providers with lb on nodes
    # kube_apiserver_port used on providers with external lb

    port = deploy_params.get('ucp_kube_apiserver_port') or deploy_params.get('kube_apiserver_port')
    assert port, "Kube API port not found in LCMCluster task 'deploy' for 'control' nodes"
    secret_name = f"{cluster_name}-kubeconfig"
    user_cert_data, cert_auth_data, client_cert_data = get_cert_data_from_kubeconfig()
    show_step(4)
    bootstrap.generate_cert(f'mke-{cluster_name}', ip=ip, dns_name=dns_name, cluster_name=cluster_name)
    bootstrap.set_cert('mke', f'mke-{cluster_name}', cluster_name=cluster_name,
                       cluster_ns=namespace_name, ip=ip, dns_name=dns_name)
    cert = bootstrap.get_cert_from_seed(cluster_name)
    new_certificate = x509.load_pem_x509_certificate(cert.encode())
    show_step(5)
    waiters.wait(
        lambda: get_cert(host) == new_certificate,
        interval=10, timeout=300,
        timeout_msg="Certificate not changed")

    show_step(6)
    # WR for PRODX-27128
    waiters.wait(
        lambda: is_kubeconfig_changed(user_cert_data, cert_auth_data, client_cert_data),
        interval=10, timeout=300,
        timeout_msg="Kubeconfig was not updated with new cert data.")

    show_step(7)
    # renew cluster obj with updated cert
    cluster.check.check_cert_conversation(host)
    cluster.check.check_cert_equal(port, cert)
    cluster.store_k8s_artifacts()
