from si_tests import logger
from kombu import Connection
LOG = logger.logger


def test_external_rabbitmq_notifications(os_manager, kaas_manager):
    """Check RabbitMQ external notifications
    1) Check enabled external notifications
    2) Check RabbitMQ service exposed outside
    3) Get credentials from the external secret
    4) Comparing parameters from secret and service
    5) Check connect to RabbitMQ vhost
       Parameters required for test execution:
         - KUBECONFIG
    """
    ext_rabbitmq_service = 'rabbitmq-external'
    ext_secrets_namespace = 'openstack-external'

    LOG.info("Check enabled external notifications")
    osdpl = os_manager.get_osdpl_deployment()
    enabled_ext_notifications = (
        osdpl.data.get('spec')
        .get('features', {})
        .get('messaging', {})
        .get('notifications', {})
        .get('external', {})
        .get("enabled", False)
    )
    ext_topics = (
        osdpl.data.get('spec')
        .get('features', {})
        .get('messaging', {})
        .get('notifications', {})
        .get('external', {})
        .get("topics", [])
    )

    assert enabled_ext_notifications, "RabbitMQ external notifications aren't enabled"

    assert ext_topics, "RabbitMQ doesn't have external topics"

    LOG.info("Get rabbitmq external service")
    rabbitmq_ext = os_manager.get_os_service(ext_rabbitmq_service)
    LOG.info("Rabbitmq external service enabled")
    ext_ip_rabbitmq = rabbitmq_ext.get_external_ip()
    ports = rabbitmq_ext.get_ports()
    for port in ports:
        if port.name == 'amqp':
            ext_port_rabbitmq = str(port.port)
    LOG.info(f"Rabbitmq external service has external ip: {ext_ip_rabbitmq}, port: {ext_port_rabbitmq}")

    for ext_topic in ext_topics:
        LOG.info(f'Get credentials from the external secret for topic: {ext_topic}')
        secret_data = kaas_manager.get_secret_all_data(f'openstack-{ext_topic}-notifications', ext_secrets_namespace)
        LOG.info(f'Got credentials for access to external RabbitMQ queue: {secret_data}')

        LOG.info("Comparing parameters from secret and service")
        if ext_ip_rabbitmq != secret_data["hosts"] or ext_port_rabbitmq != secret_data["port_amqp"]:
            LOG.error(f'Comparison error, values are different.'
                      f'Service: {ext_ip_rabbitmq}, {ext_port_rabbitmq}.'
                      f'Secret: {secret_data["hosts"]}, {secret_data["port_amqp"]}')
            raise Exception("Comparison error, values are different.")

        LOG.info(f'Check connect to RabbitMQ vhost {secret_data["vhost"]}')
        check_test_connection = check_connection(
            secret_data["username"],
            secret_data["password"],
            secret_data["hosts"],
            secret_data["port_amqp"],
            secret_data["vhost"],
        )
        if check_test_connection:
            LOG.info("The connection check to the RabbitMQ virtual host is complete.")

        assert check_test_connection, "Failed connection to RabbitMQ vhost"


def check_connection(username, password, host, port, vhost):
    rabbitmq_url = f'amqp://{username}:{password}@{host}:{port}/{vhost}'
    connection = Connection(rabbitmq_url)
    try:
        LOG.info(f"Connecting to the: {rabbitmq_url}")
        connection.ensure_connection(max_retries=3)
        connection.channel()
    except Exception as e:
        LOG.error(f"Connection error. Error: {e}")
        return False
    finally:
        connection.release()
    return True
