import pytest

import yaml
import json

from kubernetes.client import V1Namespace

from si_tests import logger
from si_tests import settings
from si_tests.fixtures.kubectl import create_secret
from si_tests.utils import templates


LOG = logger.logger


def run_tests(secret_name, kubectl_client, name=settings.RALLY_POD_NAME):
    """Run rally tests"""
    ns = settings.RALLY_NAMESPACE

    options = {
        'RALLY_POD_NAME': name,
        'KUBECONFIG_SECRET_NAME': secret_name
    }

    if not settings.RALLY_ON_DEDICATED_CLUSTER:
        template = templates.render_template(
            settings.RALLY_POD_YAML_WITHOUT_DEDICATED_CLUSTER, options)
    else:
        template = templates.render_template(
            settings.RALLY_POD_YAML, options)
    LOG.debug(template)
    json_body = json.dumps(yaml.load(template, Loader=yaml.SafeLoader))

    if kubectl_client.pods.present(name=name, namespace=ns):
        LOG.warning("Pod '{0}' already exists. Deleting..".format(name))
        # kubectl_client.api_core.delete_namespaced_pod(name, ns)
        kubectl_client.pods.get(name=name, namespace=ns).delete()

    pod = kubectl_client.pods.create(name=name,
                                     namespace=ns,
                                     body=json.loads(json_body))
    pod.wait_phase('Running', timeout=60, interval=10)
    return pod


def wait_test_results(pod, timeout=settings.RALLY_WAIT_TIMEOUT):
    pod.wait_test(filepath='./test_*', timeout=timeout)


def check_test_results(pod):
    """Check if the file ./test_passed presents in the pod

    Return True if present, False otherwise
    """
    command = ['/bin/sh', '-c', 'test -f ./test_passed && echo "OK"']
    return pod.exec(command) == "OK\n"


def download_test_results(pod):
    pod.cp_from_pod(source_dir='/rally-reports')


def _get_rally_node(nodesList):
    for node in nodesList:
        node_labels = node.data['metadata']['labels']
        if node_labels.get(settings.RALLY_LABEL_KEY) == settings.RALLY_LABEL_VALUE:
            return True, node
    return False, nodesList[0]


def add_label_daemonset_node(kubectl_client):
    label_selector_daemonset = ('stacklight!=enabled,!node-role.kubernetes.io/master,' +
                                '{0}!={1}'.format(settings.RALLY_LABEL_KEY, settings.RALLY_LABEL_VALUE))
    nodes_daemonset = kubectl_client.nodes.list(label_selector=label_selector_daemonset)
    label_daemonset = {'rally': 'daemonset'}
    for node in nodes_daemonset:
        node.add_label(label_daemonset)


def prepare_rally_node(kubectl_client):
    taint = {"key": settings.RALLY_TOLERATION_KEY,
             "value": settings.RALLY_TOLERATION_VALUE,
             "effect": settings.RALLY_TOLERATION_EFFECT}
    label_selector = 'stacklight!=enabled,!node-role.kubernetes.io/master'
    nodes = kubectl_client.nodes.list(label_selector=label_selector)
    exist_rally_label, rally_node = _get_rally_node(nodes)
    label = {settings.RALLY_LABEL_KEY: settings.RALLY_LABEL_VALUE}
    if not exist_rally_label:
        rally_node.add_label(label)
    if not rally_node.exists_taint(taint):
        rally_node.add_taint(taint)


def create_kubeconfig_secret(kubectl_client, content, namespace, name):
    config = ''
    if settings.INTERNAL_K8S_ENDPOINT_ENABLED:
        INTERNAL_K8S_ENDPOINT = 'https://kubernetes.default.svc.cluster.local:443'
        kubeconfig = yaml.load(content, Loader=yaml.SafeLoader)
        for cluster in kubeconfig['clusters']:
            cluster['cluster']['server'] = INTERNAL_K8S_ENDPOINT
        config = yaml.dump(kubeconfig)
    else:
        config = content
    create_secret(kubectl_client, config, namespace, name)


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
def test_k8s_rally(kaas_manager, _):
    # Collect data and prepare for test
    if settings.RALLY_ON_DEDICATED_CLUSTER:
        # cluster with rally:
        # 1. get testcluster k8s client
        ts_name = settings.TESTCLUSTER_NAME
        ts_ns = kaas_manager.get_namespace(settings.TESTCLUSTER_NAMESPACE)
        ts_cluster = ts_ns.get_cluster(ts_name)
        ts_client = ts_cluster.k8sclient
        # target cluster:
        # 2. get target (tested) cluster kubeconfig
        tg_name = settings.TARGET_CLUSTER
        tg_ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
        tg_cluster = tg_ns.get_cluster(tg_name)
        _, tg_kubeconfig = tg_cluster.get_kubeconfig_from_secret()
    else:
        ts_name = tg_name = settings.TARGET_CLUSTER
        ts_ns = tg_ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
        ts_cluster = tg_cluster = ts_ns.get_cluster(ts_name)
        ts_client = ts_cluster.k8sclient
        _, tg_kubeconfig = tg_cluster.get_kubeconfig_from_secret()
    # 3. create ns in target cluster
    rally_ns = settings.RALLY_NAMESPACE
    if not ts_client.namespaces.present(name=rally_ns):
        LOG.info("Creating Namespace '{0}'".format(rally_ns))
        ns_body = V1Namespace(api_version='v1',
                              kind='Namespace',
                              metadata={'name': rally_ns})
        ts_client.api_core.create_namespace(body=ns_body)
    else:
        LOG.info("Namespace '{0}' already exists".format(rally_ns))
    # 4. create secret in target cluster
    LOG.debug("Creating target cluster kubeconfig secret")
    secret_name = '-'.join(('kubeconfig', tg_name))
    if ts_client.secrets.present(name=secret_name, namespace=rally_ns):
        LOG.warning("Secret '{0}' already exists. Deleting.."
                    .format(secret_name))
        ts_client.api_core.delete_namespaced_secret(
            secret_name, rally_ns
        )
    create_kubeconfig_secret(ts_client, tg_kubeconfig, rally_ns, secret_name)

    # Run test
    if not settings.RALLY_ON_DEDICATED_CLUSTER:
        LOG.info('Adding label and taint for rally node')
        prepare_rally_node(ts_client)
    LOG.info('Adding labels for the daemonset scenario')
    add_label_daemonset_node(ts_client)
    LOG.info("Starting rally pod")
    pod = run_tests(secret_name, ts_client)
    LOG.info("Waiting for results ...")
    wait_test_results(
        pod=pod, timeout=settings.RALLY_WAIT_TIMEOUT
    )
    download_test_results(pod)
    k8s_rally_pass = check_test_results(pod)

    # Checking all pods are in consistent state after test run
    ns = kaas_manager.get_namespace(namespace=settings.TARGET_NAMESPACE)
    cluster = ns.get_cluster(settings.TARGET_CLUSTER)
    cluster.check.check_k8s_pods()

    if settings.RALLY_CHECK_RESULTS:
        assert k8s_rally_pass,  "Some of k8s rally tests failed"
